#include "surf_routing_cluster_fat_tree.hpp"
#include "xbt/lib.h"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <iostream>
#include <fstream>
#include <sstream>

XBT_LOG_NEW_DEFAULT_SUBCATEGORY(surf_route_fat_tree, surf, "Routing for fat trees");

AS_t model_fat_tree_cluster_create(void)
{
  return new AsClusterFatTree();
}

AsClusterFatTree::AsClusterFatTree() : levels(0) {
  XBT_DEBUG("Creating a new fat tree.");
}

AsClusterFatTree::~AsClusterFatTree() {
  for (unsigned int i = 0 ; i < this->nodes.size() ; i++) {
    delete this->nodes[i];
  }
}

bool AsClusterFatTree::isInSubTree(FatTreeNode *root, FatTreeNode *node) {
  for (unsigned int i = 0 ; i < node->level ; i++) {
    if(root->label[i] != node->label[i]) {
      return false;
    }
  }
  
  for (unsigned int i = root->level + 1 ; i < this->levels ; i++) {
    if(root->label[i] != node->label[i]) {
      return false;
    }
  }
  return true;
}

void AsClusterFatTree::getRouteAndLatency(RoutingEdgePtr src,
                                          RoutingEdgePtr dst,
                                          sg_platf_route_cbarg_t into,
                                          double *latency) {
  FatTreeNode *source, *destination, *currentNode;
  std::vector<NetworkLink*> route;
  source = this->computeNodes.find(src->getId())->second;
  destination = this->computeNodes.find(dst->getId())->second;

  XBT_DEBUG("Get route and latency from '%s' [%d] to '%s' [%d] in a fat tree",
            src->getName(), src->getId(), dst->getName(), dst->getId());

  currentNode = source;

  // up part
  while (!isInSubTree(currentNode, destination)) {
    int d, k; // as in d-mod-k
    d = destination->position;

    for (unsigned int i = 0 ; i < currentNode->level ; i++) {
      d /= this->upperLevelNodesNumber[i];
    }
     k = this->upperLevelNodesNumber[currentNode->level] *
      this->lowerLevelNodesNumber[currentNode->level];
     d = d % k;
     route.push_back(currentNode->parents[d]->upLink);
     if(latency) {
       *latency += currentNode->parents[d]->upLink->getLatency();
     }
     currentNode = currentNode->parents[d]->upNode;
  }
  
  // Down part
  while(currentNode != destination) {
    for(unsigned int i = 0 ; i < currentNode->children.size() ; i++) {
      if(i % this->lowerLevelNodesNumber[currentNode->level] ==
         destination->label[currentNode->level]) {
        route.push_back(currentNode->children[i]->downLink);
        if(latency) {
          *latency += currentNode->children[i]->downLink->getLatency();
        }
        currentNode = currentNode->children[i]->downNode;
      }
    }
  }
  
  for (unsigned int i = 0 ; i < route.size() ; i++) {
    xbt_dynar_push_as(into->link_list, void*, route[i]);
  }

}

/* This function makes the assumption that parse_specific_arguments() and
 * addNodes() have already been called
 */
void AsClusterFatTree::create_links(sg_platf_cluster_cbarg_t cluster){
  if(this->levels == 0) {
    return;
  }
  this->generateSwitches();


  if(XBT_LOG_ISENABLED(surf_route_fat_tree, xbt_log_priority_debug)) {
    std::stringstream msgBuffer;

    msgBuffer << "We are creating a fat tree of " << this->levels << " levels "
              << "with " << this->nodesByLevel[0] << " processing nodes";
    for (unsigned int i = 1 ; i <= this->levels ; i++) {
      msgBuffer << ", " << this->nodesByLevel[i] << " switches at level " << i;
    }
    XBT_DEBUG(msgBuffer.str().c_str());
    msgBuffer.str("");
    msgBuffer << "Nodes are : ";

    for (unsigned int i = 0 ;  i < this->nodes.size() ; i++) {
      msgBuffer << this->nodes[i]->id << "(" << this->nodes[i]->level << ","
                << this->nodes[i]->position << ") ";
    }
    XBT_DEBUG(msgBuffer.str().c_str());
  }


  this->generateLabels();

  unsigned int k = 0;
  // Nodes are totally ordered, by level and then by position, in this->nodes
  for (unsigned int i = 0 ; i < this->levels ; i++) {
    for (unsigned int j = 0 ; j < this->nodesByLevel[i] ; j++) {
        this->connectNodeToParents(cluster, this->nodes[k]);
        k++;
    }
  }
  
  if(XBT_LOG_ISENABLED(surf_route_fat_tree, xbt_log_priority_debug)) {
    std::stringstream msgBuffer;
    msgBuffer << "Links are : ";
    for (unsigned int i = 0 ; i < this->links.size() ; i++) {
      msgBuffer << "(" << this->links[i]->upNode->id << ","
                << this->links[i]->downNode->id << ") ";
    }
    XBT_DEBUG(msgBuffer.str().c_str());
  }


}

int AsClusterFatTree::connectNodeToParents(sg_platf_cluster_cbarg_t cluster,
                                           FatTreeNode *node) {
  std::vector<FatTreeNode*>::iterator currentParentNode = this->nodes.begin();
  int connectionsNumber = 0;
  const int level = node->level;
  XBT_DEBUG("We are connecting node %d(%u,%u) to his parents.",
            node->id, node->level, node->position);
  currentParentNode += this->getLevelPosition(level + 1);
  for (unsigned int i = 0 ; i < this->nodesByLevel[level + 1] ; i++ ) {
    if(this->areRelated(*currentParentNode, node)) {
      XBT_DEBUG("%d(%u,%u) and %d(%u,%u) are related,"
                " with %u links between them.", node->id,
                node->level, node->position, (*currentParentNode)->id,
                (*currentParentNode)->level, (*currentParentNode)->position, this->lowerLevelPortsNumber[level]);
      for (unsigned int j = 0 ; j < this->lowerLevelPortsNumber[level] ; j++) {
      this->addLink(cluster, *currentParentNode, node->label[level + 1] +
                    j * this->lowerLevelNodesNumber[level], node,
                    (*currentParentNode)->label[level + 1] +
                    j * this->upperLevelNodesNumber[level]);
      }
      connectionsNumber++;
    }
    ++currentParentNode;
  }
  return connectionsNumber;
}


bool AsClusterFatTree::areRelated(FatTreeNode *parent, FatTreeNode *child) {
  std::stringstream msgBuffer;

  if(XBT_LOG_ISENABLED(surf_route_fat_tree, xbt_log_priority_debug)) {
    msgBuffer << "Are " << child->id << "(" << child->level << ","
              << child->position << ") <";

    for (unsigned int i = 0 ; i < this->levels ; i++) {
      msgBuffer << child->label[i] << ",";
    }
    msgBuffer << ">";
    
    msgBuffer << " and " << parent->id << "(" << parent->level
              << "," << parent->position << ") <";
    for (unsigned int i = 0 ; i < this->levels ; i++) {
      msgBuffer << parent->label[i] << ",";
    }
    msgBuffer << ">";
    msgBuffer << " related ? ";
    XBT_DEBUG(msgBuffer.str().c_str());
    
  }
  if (parent->level != child->level + 1) {
    return false;
  }
  
  for (unsigned int i = 1 ; i <= this->levels; i++) {
    if (parent->label[this->levels - i] != child->label[this->levels - i] && i != parent->level) {
      return false;
    }
  }
  return true;
}

void AsClusterFatTree::generateSwitches() {
  XBT_DEBUG("Generating switches.");
  this->nodesByLevel.resize(this->levels + 1, 0);
  unsigned int nodesRequired = 0;

  // We take care of the number of nodes by level
  this->nodesByLevel[0] = 1;
  for (unsigned int i = 0 ; i < this->levels ; i++) {
    this->nodesByLevel[0] *= this->lowerLevelNodesNumber[i];
  }

     
  if(this->nodesByLevel[0] > this->nodes.size()) {
    surf_parse_error("There is not enough nodes to fit to the described topology."
                     " Please check your platform description (We need %d nodes, we only got %zu)",
                     this->nodesByLevel[0], this->nodes.size());
    return;
  }
  
  for (unsigned int i = 0 ; i < this->levels ; i++) {
    int nodesInThisLevel = 1;
      
    for (unsigned int j = 0 ;  j <= i ; j++) {
      nodesInThisLevel *= this->upperLevelNodesNumber[j];
    }
      
    for (unsigned int j = i+1 ; j < this->levels ; j++) {
      nodesInThisLevel *= this->lowerLevelNodesNumber[j];
    }

    this->nodesByLevel[i+1] = nodesInThisLevel;
    nodesRequired += nodesInThisLevel;
  }


  // If we have to many compute nodes, we ditch them
  if (this->nodesByLevel[0] < this->nodes.size()) {
    for (unsigned int i = this->nodesByLevel[0] ; i < this->nodes.size() ; i++) {
      this->computeNodes.erase(this->nodes[i]->id);
      delete this->nodes[i];
    }
    this->nodes.resize(this->nodesByLevel[0]);
  }

  // We create the switches
  int k = 0;
  for (unsigned int i = 0 ; i < this->levels ; i++) {
    for (unsigned int j = 0 ; j < this->nodesByLevel[i + 1] ; j++) {
      FatTreeNode* newNode;
      newNode = new FatTreeNode(--k, i + 1, j);
      XBT_DEBUG("We create the switch %d(%d,%d)", newNode->id, newNode->level, newNode->position);
      newNode->children.resize(this->lowerLevelNodesNumber[i] *
                               this->lowerLevelPortsNumber[i]);
      if (i != this->levels - 1) {
        newNode->parents.resize(this->upperLevelNodesNumber[i + 1]);
      }
      newNode->label.resize(this->levels);
      this->nodes.push_back(newNode);
    }
  }
}

void AsClusterFatTree::generateLabels() {
  XBT_DEBUG("Generating labels.");
  // TODO : check if nodesByLevel and nodes are filled
  std::vector<int> maxLabel(this->levels);
  std::vector<int> currentLabel(this->levels);
  unsigned int k = 0;
  for (unsigned int i = 0 ; i <= this->levels ; i++) {
    currentLabel.assign(this->levels, 0);
    for (unsigned int j = 1 ; j <= this->levels ; j++) {
      maxLabel[maxLabel.size() - j] = j > i ?
        this->lowerLevelNodesNumber[j - 1] : this->upperLevelNodesNumber[j - 1];
    }
    
    for (unsigned int j = 0 ; j < this->nodesByLevel[i] ; j++) {

      if(XBT_LOG_ISENABLED(surf_route_fat_tree, xbt_log_priority_debug )) {
        std::stringstream msgBuffer;

        msgBuffer << "Assigning label <";
        for (unsigned int l = 0 ; l < this->levels ; l++) {
          msgBuffer << currentLabel[l] << ",";
        }
        msgBuffer << "> to " << k << " (" << i << "," << j <<")";
        
        XBT_DEBUG(bprintf(msgBuffer.str().c_str()));
      }
      this->nodes[k]->label.assign(currentLabel.begin(), currentLabel.end());

      bool remainder = true;
      
      int pos = currentLabel.size() - 1;
      do {
        std::stringstream msgBuffer;

        ++currentLabel[pos];
        if (currentLabel[pos] >= maxLabel[pos] && pos > 0) {
          currentLabel[pos] = 0;
          remainder = true;
        }
        else {
          remainder = false;
        }
        if (!remainder) {
          pos = currentLabel.size() - 1;
        }
        else {
          --pos;
        }
      }
      while(remainder);
      k++;
    }
  }
}


int AsClusterFatTree::getLevelPosition(const unsigned  int level) {
  if (level > this->levels) {
    // Well, that should never happen. Maybe should we throw instead.
    return -1;
  }
  int tempPosition = 0;

  for (unsigned int i = 0 ; i < level ; i++) {
    tempPosition += this->nodesByLevel[i];
  }
 return tempPosition;
}

void AsClusterFatTree::addProcessingNode(int id) {
  using std::make_pair;
  static int position = 0;
  FatTreeNode* newNode;
  newNode = new FatTreeNode(id, 0, position++);
  newNode->parents.resize(this->upperLevelNodesNumber[0] * this->lowerLevelPortsNumber[0]);
  newNode->label.resize(this->levels);
  this->computeNodes.insert(make_pair(id,newNode));
  this->nodes.push_back(newNode);
}

void AsClusterFatTree::addLink(sg_platf_cluster_cbarg_t cluster, 
                               FatTreeNode *parent, unsigned int parentPort,
                               FatTreeNode *child, unsigned int childPort) {
  FatTreeLink *newLink;
  newLink = new FatTreeLink(cluster, parent, child);
  XBT_DEBUG("Creating a link between the parent (%d,%d,%u)"
            " and the child (%d,%d,%u)", parent->level, parent->position,
            parentPort, child->level, child->position, childPort);
  parent->children[parentPort] = newLink;
  child->parents[childPort] = newLink;

  this->links.push_back(newLink);

  

}

void AsClusterFatTree::parse_specific_arguments(sg_platf_cluster_cbarg_t 
                                                cluster) {
  std::vector<string> parameters;
  std::vector<string> tmp;
  boost::split(parameters, cluster->topo_parameters, boost::is_any_of(";"));
 

  // TODO : we have to check for zeros and negative numbers, or it might crash
  if (parameters.size() != 4){
    surf_parse_error("Fat trees are defined by the levels number and 3 vectors" 
                     ", see the documentation for more informations");
    // Well, there's no doc, yet
  }

  // The first parts of topo_parameters should be the levels number
  this->levels = std::atoi(parameters[0].c_str()); // stoi() only in C++11...
  
  // Then, a l-sized vector standing for the childs number by level
  boost::split(tmp, parameters[1], boost::is_any_of(","));
  if(tmp.size() != this->levels) {
    surf_parse_error("Fat trees are defined by the levels number and 3 vectors" 
                     ", see the documentation for more informations"); 
  }
  for(size_t i = 0 ; i < tmp.size() ; i++){
    this->lowerLevelNodesNumber.push_back(std::atoi(tmp[i].c_str())); 
  }
  
  // Then, a l-sized vector standing for the parents number by level
  boost::split(tmp, parameters[2], boost::is_any_of(","));
  if(tmp.size() != this->levels) {
    surf_parse_error("Fat trees are defined by the levels number and 3 vectors" 
                     ", see the documentation for more informations"); 
  }
  for(size_t i = 0 ; i < tmp.size() ; i++){
    this->upperLevelNodesNumber.push_back(std::atoi(tmp[i].c_str())); 
  }
  
  // Finally, a l-sized vector standing for the ports number with the lower level
  boost::split(tmp, parameters[3], boost::is_any_of(","));
  if(tmp.size() != this->levels) {
    surf_parse_error("Fat trees are defined by the levels number and 3 vectors" 
                     ", see the documentation for more informations"); 
    
  }
  for(size_t i = 0 ; i < tmp.size() ; i++){
    this->lowerLevelPortsNumber.push_back(std::atoi(tmp[i].c_str())); 
  }
}


void AsClusterFatTree::generateDotFile(const string& filename) const {
  ofstream file;
  /* Maybe should we get directly a char*, as open takes strings only beginning
   * with C++11...
   */
  file.open(filename.c_str(), ios::out | ios::trunc); 
  
  if(file.is_open()) {
    file << "graph AsClusterFatTree {\n";
    for (unsigned int i = 0 ; i < this->nodes.size() ; i++) {
      file << this->nodes[i]->id;
      if(this->nodes[i]->id < 0) {
        file << " [shape=circle];\n";
      }
      else {
        file << " [shape=hexagon];\n";
      }
    }

    for (unsigned int i = 0 ; i < this->links.size() ; i++ ) {
      file << this->links[i]->downNode->id
             << " -- "
           << this->links[i]->upNode->id
             << ";\n";
    }
    file << "}";
    file.close();
  }
  else {
    std::cerr << "Unable to open file " << filename << std::endl;
    return;
  }
}

FatTreeNode::FatTreeNode(int id, int level, int position) : id(id),
                                                            level(level),
                                                            position(position){}

FatTreeLink::FatTreeLink(sg_platf_cluster_cbarg_t cluster, FatTreeNode *downNode,
                         FatTreeNode *upNode) : upNode(upNode),
                                                downNode(downNode) {
  static int uniqueId = 0;
  s_sg_platf_link_cbarg_t linkTemplate;
  memset(&linkTemplate, 0, sizeof(linkTemplate));
  linkTemplate.bandwidth = cluster->bw;
  linkTemplate.latency = cluster->lat;
  linkTemplate.state = SURF_RESOURCE_ON;
  linkTemplate.policy = cluster->sharing_policy; // Maybe should we do sthg with that ?
  linkTemplate.id = bprintf("link_from_%d_to_%d_%d", downNode->id, upNode->id, uniqueId);
  sg_platf_new_link(&linkTemplate);
  NetworkLink* link;
  if (cluster->sharing_policy == SURF_LINK_FULLDUPLEX) {
    std::string tmpID;
    tmpID = std::string(linkTemplate.id) + "_UP";
    link = (NetworkLink*) xbt_lib_get_or_null(link_lib, tmpID.c_str(), SURF_LINK_LEVEL);
    this->upLink = link; // check link?
    tmpID = std::string(linkTemplate.id) + "_DOWN";
    link = (NetworkLink*) xbt_lib_get_or_null(link_lib, tmpID.c_str(), SURF_LINK_LEVEL);
    this->downLink = link; // check link ?
  }
  else {
    link = (NetworkLink*) xbt_lib_get_or_null(link_lib, linkTemplate.id, SURF_LINK_LEVEL);
    this->upLink = link;
    this->downLink = link;
  }
  uniqueId++;
  
}
